import copy
from typing import Union
import requests
import webbrowser
from typing import List

import pandas as pd
import pyarrow as pa

from ..._common.config import *
from ..._common.config import URL_TASK, URL_UNIVERSES, URL_UPLOAD
from ..._common.const import (
    FBFrequencyType as _FrequencyType,
    RankType as _RankType,
    CurrencyTypeWithReportTrade as _CurrencyTypeWithReportTrade
)
from ..._core._req_builder._universe import parse_universe_to_universeid
from ..._core._req_builder._portfolio import parse_portfolios_to_portfolioids
from ..._core._req_builder._portfolio import should_overwrite_portfolio
from ..._utils import (
    _authentication,
    get,
    _process_fileresponse,
    _validate_args,
    _get_web_authentication_token,
    are_periods_exclusive,
    Loader, post, get, are_periods_exclusive,
)
from ..._utils.exceptions import PrismTaskError, PrismValueError
from ..._prismcomponent import abstract_prismcomponent
from ..._prismcomponent.prismcomponent import _PrismTaskComponent

_data_category = __name__.split(".")[-1]

class PrismBacktestTaskComponent(_PrismTaskComponent):
    _component_category_repr = _data_category

    @classmethod
    def _get_task_result(cls, tasktype, tasktyperepr, resultid: list, data: bool = True, report: bool = True):
        if (data == False) and (report == False):
            raise PrismValueError("Either data or report should be true.")
        if report:
            # Format Definition:
            # Webclient Address/report/report_type/file_id?token=web_auth_token
            url_get_results = [f"{URL_TASK}/{cls.componentid}/report/{i}" for i in resultid]
            fileids = [get(url)["path"].values[0] for url in url_get_results]

            web_auth_token = _get_web_authentication_token()

            print("Fetching link to " + tasktyperepr + " Report...")
            for idx, i in enumerate(fileids):
                link = f"{ROOT_EXT_WEB_URL}/report/{cls.__name__}/{i}"
                print(f"Link to {tasktyperepr} Report {resultid[idx]}:")
                print(link)

            if len(resultid) == 1:
                link = link[:-1] + "?token=" + web_auth_token
                webbrowser.open(link, new=2)

        if data:
            url_get_results = [f"{URL_TASK}/{cls.componentid}/result/{i}" for i in resultid]
            print("Fetching " + tasktyperepr + " Result Data...")
            headers = _authentication()
            ret = {}
            for idx, i in enumerate(resultid):
                res = requests.get(url_get_results[idx], headers=headers)
                ret_dict, _ = _process_fileresponse(res, res.content)
                ret[i] = {k.split(".parquet")[0]: v for k, v in ret_dict.items()}
            return list(ret.values())[0] if len(ret) == 1 else ret


class factor_backtest(PrismBacktestTaskComponent):
    """
    Enables users to quickly test and identify factors which may predict future return.

    Parameters
    ----------
        jobname : str
            | Name of the job when the task component is run.
            | If None, the default job name sets to factorbacktest_{jobid}.

        bins : int
            | Number of quantile portfolio the universe generate. Should be bigger than 1 and smaller than or equal to 20.
            | If specified, this will overwrite bins parameter in the task component.

            .. admonition: Note
                :class: note

                Bins are assigned to the factor score in descending order. Meaning the largest factor score will be assigned 1st bin

        frequency : str
            | Desired sampling frequency to resample.
            |
            | Frequency Format
            | Description: The 'frequency format' combines all the elements of the 'frequency' parameter, providing a comprehensive way to specify the resampling frequency.
            | Format: XBF-S
            |
            | Elements of the frequency Parameter
            | X: Resampling Interval (Optional, If not given, it is defaults to be 1, meaning resampling occurs every data point)
            | Description: The 'X' element represents the number at which resampling should be operated. It determines how often data should be resampled.
            |
            | B: Business Frequency (Optional, If not given, it is assumed to be a non-business frequency)
            | Description: The 'B' element indicates whether the resampling frequency should align with business days. If 'B' is included, it signifies a business frequency; otherwise, it's not considered.
            |
            | F: Frequency Type
            | Description: The 'F' element specifies the frequency type, which defines the unit of time for resampling. It can take on one of the following values: [D, W, M, Q, SA, A].
            |
            | S: Specific Part of Frequency (Optional)
            | Description: The 'S' element is an optional part of the 'frequency' parameter, and its presence depends on the chosen frequency type ('F'). It allows you to specify a particular aspect of the resampling frequency.
            |

                | Specific Part for Each Frequency Type
                | A (Annual) Frequency - MM/dd (Months and Day)
                |    Example: A-12/15 (Resample every year on December 15th)
                |
                | M (Monthly), Q (Quarterly), SA (Semi-Annual) Frequencies - dd (Day of the Month)
                |    Example: 3M-25 (Resample every 3 months on the 25th day of the month)
                |
                | W (Weekly) Frequency - Day of the Week (Case Insensitive)
                |    Example: 2W-Fri (Resample every 2 weeks on Fridays)
                |
                | D (Daily) Frequency - N/A (If specific part is given it will raise an error)

            .. admonition:: Note
                :class: note

                | Result for certain frequencies
                |
                | For Q (Quarterly) frequency, the resampled results can only be for the months of March, June, September, and December.
                | For SA (Semi-Annual) frequency, the resampled results can only be for the months of June and December.
                | For dynamic months, use 3M instead of Q and 6M instead of SA.

            | Example
            - 3M-25: Resample every 3 months on the 25th day of the month.
            - 3D: Resample every 3 days.
            - 2BM-15: Resample every 2 months on the 15th day of the month, considering business days.
            - A-12/15: Resample every year on December 15th.
            - 2BQ-3: Resample every 2 quarters, on the 3rd day of the month, considering business days.

        lookback : int
            The periods to lookback are defined by the resampling frequency parameter. For example, if resampling to Monthly data, this will lookback *lookback* Months.

            .. admonition:: Note
                :class: note

                | When up-sampling, the lookback input parameter must be specified properly.
                |
                | For example, if resampling from Quarterly data to Monthly data, the lookback should be at least '3' or larger.
                | If set to only '1', then the lookback will not look far enough back to fill in every month and missing values will be left in the output time-series.
                | If no input is supplied, it will go back to the last data available and fill in every missing value in between.


        startdate : str, default None
            | Startdate of the time period for which to load data or the window in time in which to run a task.
            | If specified, this will overwrite startdate parameter in the task component.

        enddate : str, default None
            | Enddate of the time period for which to load data or the window in time in which to run a task.
            | If specified, this will overwrite enddate parameter in the task component.

        max_days : int, default None
            | If None, default max days is induced from rebalancing frequency.
            | If specified, this will overwrite max_days parameter in the task component.

        rank_method : str {'standard', 'modified', 'dense', 'ordinal', 'fractional'}, default 'standard'
            | Method for how equal values are assigned a rank.

            - standard : 1 2 2 4
            - modified : 1 3 3 4
            - dense : 1 2 2 3
            - ordinal : 1 2 3 4
            - fractional : 1 2.5 2.5 4

            | Desired rebalancing frequency to run screen.
            | If specified, this will overwrite rank_method parameter in the task component.

    Returns
    -------
        status : dict
            Status of factorbacktest run.

    Examples
    --------
        >>> ni = prism.financial.income_statement(dataitemid=100639, periodtype='LTM')
        >>> mcap = prism.market.market_cap()
        >>> ep = ni / mcap
        >>> fb_ep = prism.factor_backtest(
                factor=ep,
                universe='Russell 3000 Index',
                frequency='Q',
                bins=5,
                startdate='2010-01-01',
                enddate='2015-01-01'
                )
        >>> fb_price_mom.run(jobname="factor_backtest_ep")
        factor_backtest is added to worker queue!: jobid is 1

        >>> prism.job_manager()
        >>> # Wait for the job 1 in GUI until its status changed to 'Completed'

        >>> prism.factor_backtest.get_result(1)
        Done!
        factor_backtest Completed: factorbacktestid is 1
        Fetching A Link to Factor Backtest Report...
        Link to Factor Backtest Report:
        https://ext.prism39.com/report/factor_backtest/my_username_1_afc7730c-55e6-41a8-ad4f-77df20caecc9/
    """
    @_validate_args
    def __init__(
        self,
        factor: Union[abstract_prismcomponent._AbstractPrismComponent, pd.DataFrame],
        universe: Union[int, str],
        frequency: _FrequencyType,
        bins: int,
        startdate: str = None,
        enddate: str = None,
        max_days: int = None,
        rank_method: _RankType = "standard",
    ):
        if (bins < 2) or (bins > 20):
            PrismValueError("The number of bins should be between 2 and 20")

        if not isinstance(factor, list):
            factor = [factor]
        factor_ = []
        for f in factor:
            if isinstance(f, pd.DataFrame):
                if len(set(f.columns) & {"date", "listingid", "value"}) != 3:
                    raise PrismValueError("Columns should be: date, listingid, value")
                if f.empty:
                    raise PrismValueError("Dataframe should not be empty")
                factor_.append(f)
            elif isinstance(f, abstract_prismcomponent._AbstractPrismComponent):
                factor_.append(f._query)
            else:
                raise PrismValueError("Factor should be either prism component or a pandas dataframe")

        universeid, _ = parse_universe_to_universeid(universe)

        universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

        if universe_period_violated:
            raise PrismValueError(
                f'Factor Backtest period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        return super().__init__(
            factor_dataquery=factor_,
            universeid=int(universeid),
            frequency=frequency,
            bins=bins,
            rank_method=rank_method,
            max_days=max_days,
            startdate=startdate,
            enddate=enddate,
        )

    @_validate_args
    def run(
        self,
        jobname: Union[str, list] = None,
        frequency: _FrequencyType = None,
        bins: int = None,
        rank_method: str = None,
        max_days: int = None,
        startdate: str = None,
        enddate: str = None,
    ):
        """
        Enables users to quickly test and identify factors which may predict future return.

        Parameters
        ----------
            jobname : str
                | Name of the job when the task component is run.
                | If None, the default job name sets to factorbacktest_{jobid}.

            frequency : str
                | Desired sampling frequency to resample.
                |
                | Frequency Format
                | Description: The 'frequency format' combines all the elements of the 'frequency' parameter, providing a comprehensive way to specify the resampling frequency.
                | Format: XBF-S
                |
                | Elements of the frequency Parameter
                | X: Resampling Interval (Optional, If not given, it is defaults to be 1, meaning resampling occurs every data point)
                | Description: The 'X' element represents the number at which resampling should be operated. It determines how often data should be resampled.
                |
                | B: Business Frequency (Optional, If not given, it is assumed to be a non-business frequency)
                | Description: The 'B' element indicates whether the resampling frequency should align with business days. If 'B' is included, it signifies a business frequency; otherwise, it's not considered.
                |
                | F: Frequency Type
                | Description: The 'F' element specifies the frequency type, which defines the unit of time for resampling. It can take on one of the following values: [D, W, M, Q, SA, A].
                |
                | S: Specific Part of Frequency (Optional)
                | Description: The 'S' element is an optional part of the 'frequency' parameter, and its presence depends on the chosen frequency type ('F'). It allows you to specify a particular aspect of the resampling frequency.
                |

                    | Specific Part for Each Frequency Type
                    | A (Annual) Frequency - MM/dd (Months and Day)
                    |    Example: A-12/15 (Resample every year on December 15th)
                    |
                    | M (Monthly), Q (Quarterly), SA (Semi-Annual) Frequencies - dd (Day of the Month)
                    |    Example: 3M-25 (Resample every 3 months on the 25th day of the month)
                    |
                    | W (Weekly) Frequency - Day of the Week (Case Insensitive)
                    |    Example: 2W-Fri (Resample every 2 weeks on Fridays)
                    |
                    | D (Daily) Frequency - N/A (If specific part is given it will raise an error)

                .. admonition:: Note
                    :class: note

                    | Result for certain frequencies
                    |
                    | For Q (Quarterly) frequency, the resampled results can only be for the months of March, June, September, and December.
                    | For SA (Semi-Annual) frequency, the resampled results can only be for the months of June and December.
                    | For dynamic months, use 3M instead of Q and 6M instead of SA.

                | Example
                - 3M-25: Resample every 3 months on the 25th day of the month.
                - 3D: Resample every 3 days.
                - 2BM-15: Resample every 2 months on the 15th day of the month, considering business days.
                - A-12/15: Resample every year on December 15th.
                - 2BQ-3: Resample every 2 quarters, on the 3rd day of the month, considering business days.

            lookback : int
                The periods to lookback are defined by the resampling frequency parameter. For example, if resampling to Monthly data, this will lookback *lookback* Months.

                .. admonition:: Note
                    :class: note

                    | When up-sampling, the lookback input parameter must be specified properly.
                    |
                    | For example, if resampling from Quarterly data to Monthly data, the lookback should be at least '3' or larger.
                    | If set to only '1', then the lookback will not look far enough back to fill in every month and missing values will be left in the output time-series.
                    | If no input is supplied, it will go back to the last data available and fill in every missing value in between.


            bins : int
                | Number of quantile portfolio the universe generate. Should be bigger than 1 and smaller than or equal to 20.
                | If specified, this will overwrite bins parameter in the task component.

                .. admonition: Note
                    :class: note

                    Bins are assigned to the factor score in descending order. Meaning the largest factor score will be assigned 1st bin

            startdate : str, default None
                | Startdate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite startdate parameter in the task component.

            enddate : str, default None
                | Enddate of the time period for which to load data or the window in time in which to run a task.
                | If specified, this will overwrite enddate parameter in the task component.

            max_days : int, default None
                | If None, default max days is induced from rebalancing frequency.
                | If specified, this will overwrite max_days parameter in the task component.

            rank_method : str {'standard', 'modified', 'dense', 'ordinal', 'fractional'}, default 'standard'
                | Method for how equal values are assigned a rank.

                - standard : 1 2 2 4
                - modified : 1 3 3 4
                - dense : 1 2 2 3
                - ordinal : 1 2 3 4
                - fractional : 1 2.5 2.5 4

                | Desired rebalancing frequency to run screen.
                | If specified, this will overwrite rank_method parameter in the task component.

        Returns
        -------
            status : dict
                Status of factorbacktest run.

        Examples
        --------
            >>> ni = prism.financial.income_statement(dataitemid=100639, periodtype='LTM')
            >>> mcap = prism.market.market_cap()
            >>> ep = ni / mcap
            >>> fb_ep = prism.factor_backtest(
                    factor=ep,
                    universe='Russell 3000 Index',
                    frequency='Q',
                    bins=5,
                    startdate='2010-01-01',
                    enddate='2015-01-01'
                    )
            >>> fb_price_mom.run(jobname="factor_backtest_ep")
            factor_backtest is added to worker queue!: jobid is 1

            >>> prism.job_manager()
            >>> # Wait for the job 1 in GUI until its status changed to 'Completed'

            >>> prism.factor_backtest.get_result(1)
            Done!
            factor_backtest Completed: factorbacktestid is 1
            Fetching A Link to Factor Backtest Report...
            Link to Factor Backtest Report:
            https://ext.prism39.com/report/factor_backtest/my_username_1_afc7730c-55e6-41a8-ad4f-77df20caecc9/
        """
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        component_args["universeid"] = int(universeid)

        universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        if frequency is not None:
            component_args["frequency"] = frequency
        if bins is not None:
            component_args["bins"] = bins
        if rank_method is not None:
            component_args["rank_method"] = rank_method
        if max_days is not None:
            component_args["max_days"] = max_days
        if startdate is not None:
            component_args["startdate"] = startdate
        if enddate is not None:
            component_args["enddate"] = enddate

        universe_period_violated = are_periods_exclusive(
            universe_startdate, universe_enddate, component_args.get("startdate"), component_args.get("enddate")
        )

        if universe_period_violated:
            raise PrismValueError(
                f'Factor Backtest period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        if (component_args["bins"] < 2) or (component_args["bins"] > 20):
            PrismValueError("The number of bins should be between 2 and 20")
        for i in range(len(component_args["factor_dataquery"])):
            if isinstance(component_args["factor_dataquery"][i], pd.DataFrame):
                headers = {"Authorization": _authentication()["Authorization"], "client": "python"}
                batch = pa.record_batch(component_args["factor_dataquery"][i])
                sink = pa.BufferOutputStream()
                with pa.ipc.new_stream(sink, batch.schema) as writer:
                    writer.write_batch(batch)
                res = requests.post(URL_UPLOAD, files={"file": sink.getvalue()}, headers=headers)
                if res.ok:
                    path = res.json()["rescontent"]["data"]["url"]

                component_args["factor_dataquery"][i] = path
        query = {
            "component_type": self._query["component_type"],
            "componentid": self._query["componentid"],
            "component_args": component_args,
        }
        custom_data = [isinstance(f, str) for f in component_args["factor_dataquery"]]

        rescontent = None
        with Loader("Factor Backtest Running... ") as l:
            try:
                rescontent = post(
                    f"{URL_TASK}/{self.componentid}",
                    params={"jobname": jobname, "custom_data": custom_data},
                    body=query,
                )
            except:
                l.stop()
                raise PrismTaskError("Factor Backtest has failed.")
            if rescontent["status"] != "Pending":
                l.stop()
                raise PrismTaskError("Factor Backtest has failed.")

        print(
            f'{rescontent["message"]}: {rescontent["result"][0]["resulttype"]} is {rescontent["result"][0]["resultvalue"]}'
        )

    @classmethod
    @_validate_args
    def get_result(cls, fbid: Union[list, int], data=True, report=False):
        """
        Return factor backtested result.

        Parameters
        ----------
            fbid: int
                | Specify the factor backtest id.

            data: bool, default True
                | Include dataframe in returned value.

            report: bool, default False
                | Open interactive GUI report in web browser at the end of the process.

                .. admonition:: Warning
                    :class: warning

                    Either data or report should be True.

        Returns
        -------
            report = True
                Open interactive Factor Backtest Report.

            data = True
                | data : dictionary of dataframes

                - *summary: summary of factorbacktest job*
                - *ar: annual return*
                - *counts: number of securities in each bin*
                - *ic: information coefficient*
                - *pnl: profit & losses*
                - *qr: quantile return*
                - *to: turnover*


        Examples
        --------
            >>> prism.factor_backtest.list_job()
            jobid            jobname  jobstatus  ...  factorbacktestid  avg_turnover    avg_ic  top_bottom_spread  frequency  bins  max_days  rank_method  description                   period
            0      1  factor_backtest_1  Completed  ...                 1      0.475282  0.017800          -0.000383          Q  10.0      93.0     standard         None  2013-01-01 ~ 2015-01-01

            >>> prism.factor_backtest.get_result(1, report=True)
            Fetching A Link to Factor Backtest Report...
            Link to Factor Backtest Report:
            https://ext.prism39.com/report/factor_backtest/my_username_1_14798b70-4a7a-4606-8179-44f6932f34e6/
            Fetching Factor Backtest Result Data...
            {
                'ar':
                    Top-Bottom Spread     Bin 1     Bin 2     Bin 3     Bin 4     Bin 5     Bin 6     Bin 7    Bin 8     Bin 9    Bin 10
                    0          -0.000383  0.002693  0.001911  0.001745  0.001602  0.001693  0.001987  0.001952  0.00193  0.002166  0.002148,
                'counts':
                                date  Bin 1  Bin 2  Bin 3  Bin 4  Bin 5  Bin 6  Bin 7  Bin 8  Bin 9  Bin 10
                        0  2013-03-31     49     50     50     50     50     49     50     50     50      50
                        1  2013-06-30     49     49     50     49     50     49     49     50     49      50
                        2  2013-09-30     49     50     50     49     50     50     49     50     50      50
                        3  2013-12-31     49     49     49     49     49     49     49     49     49      50
                        4  2014-03-31     49     50     50     50     50     50     50     50     50      50
                        5  2014-06-30     50     50     50     50     50     50     50     50     50      50
                        6  2014-09-30     49     50     50     50     50     49     50     50     50      50
                        7  2014-12-31     49     50     50     50     50     49     50     50     50      50,
                'ic':        date         ic
                    0  2013-03-31   0.109810
                    1  2013-06-30  -0.024280
                    2  2013-09-30   0.073506
                    3  2013-12-31  -0.007544
                    4  2014-03-31  -0.034033
                    5  2014-06-30  -0.035701
                    6  2014-09-30   0.042841,
                'pnl':       date  Top-Bottom Spread     Bin 1     Bin 2     Bin 3     Bin 4     Bin 5     Bin 6     Bin 7     Bin 8     Bin 9    Bin 10
                    0  2013-03-31          -0.026218  0.107641  0.025149  0.020147  0.029949  0.025136  0.030979  0.047745  0.050728  0.053055  0.081424
                    1  2013-06-30          -0.040646  0.198980  0.095820  0.069968  0.070595  0.093182  0.100881  0.102511  0.113004  0.111119  0.154577
                    2  2013-09-30          -0.017259  0.327730  0.210149  0.150424  0.158664  0.203366  0.209441  0.189935  0.221166  0.229827  0.306704
                    3  2013-12-31          -0.041429  0.406996  0.233885  0.190352  0.175839  0.225944  0.246494  0.231905  0.218980  0.286169  0.352579
                    4  2014-03-31          -0.077449  0.530770  0.313009  0.254870  0.254214  0.268607  0.292174  0.316257  0.289976  0.334870  0.420739
                    5  2014-06-30          -0.097823  0.533991  0.319311  0.267409  0.233913  0.252408  0.294491  0.305149  0.266159  0.332309  0.392353
                    6  2014-09-30          -0.076550  0.539356  0.382519  0.349352  0.320580  0.338911  0.397894  0.390694  0.386452  0.433719  0.430054,
                'qr':        date  Top-Bottom Spread     Bin 1     Bin 2     Bin 3      Bin 4      Bin 5     Bin 6      Bin 7      Bin 8      Bin 9     Bin 10
                    0  2013-03-31          -0.026218  0.107641  0.025149  0.020147   0.029949   0.025136  0.030979   0.047745   0.050728   0.053055   0.081424
                    1  2013-06-30          -0.014817  0.082462  0.068937  0.048838   0.039464   0.066378  0.067802   0.052270   0.059269   0.055138   0.067646
                    2  2013-09-30           0.024377  0.107383  0.104332  0.075195   0.082261   0.100792  0.098611   0.079295   0.097180   0.106836   0.131760
                    3  2013-12-31          -0.024594  0.059701  0.019615  0.034707   0.014824   0.018762  0.030637   0.035272  -0.001790   0.045813   0.035107
                    4  2014-03-31          -0.037578  0.087971  0.064125  0.054201   0.066654   0.034799  0.036646   0.068472   0.058242   0.037865   0.050393
                    5  2014-06-30          -0.022083  0.002104  0.004800  0.009992  -0.016187  -0.012769  0.001793  -0.008439  -0.018463  -0.001919  -0.019980
                    6  2014-09-30           0.023579  0.003497  0.047910  0.064654   0.070238   0.069069  0.079880   0.065544   0.095006   0.076116   0.027077,
                'summary':    username  universename   startdate     enddate frequency  bins  avg_turnover  avg_ic  top_bottom_spread
                        0  my_username       S&P 500  2013-01-01  2015-01-01         Q    10      0.475282  0.0178          -0.000383,
                'to':       date  turnover
                    0 2013-03-31  0.437751
                    1 2013-06-30  0.459514
                    2 2013-09-30  0.476861
                    3 2013-12-31  0.549898
                    4 2014-03-31  0.428858
                    5 2014-06-30  0.452000
                    6 2014-09-30  0.522088
            }
        """
        if isinstance(fbid, int):
            fbid = [fbid]
        return cls._get_task_result(cls.__class__.__name__, "Factor Backtest", fbid, data, report)

    @classmethod
    def list_job(cls):
        """
        List all factor backtest jobs and their detail.

        Returns
        -------
            pandas.DataFrame
                All factor backtest jobs.
            Columns
                - *jobid*
                - *jobname*
                - *jobstatus*
                - *starttime*
                - *endtime*
                - *universeid*
                - *universepath*
                - *factorbacktestid*
                - *avg_turnover*
                - *avg_ic*
                - *top_bottom_spread*
                - *frequency*
                - *bins*
                - *max_days*
                - *rank_method*
                - *description*
                - *period*

        Examples
        --------
        >>> prism.factor_backtest.list_job()
        jobid  jobstatus                   starttime                     endtime  ...  factorbacktestid  avg_turnover     avg_ic  top_bottom_spread  frequency  bins                   period
        0	   4  Completed  2022-06-29 17:48:41.023456  2022-06-29 17:52:03.637405  ...                 4      0.854072  -0.016881          -0.000389          Q    10  2000-01-01 ~ 2020-01-01
        1      3  Completed  2022-06-29 10:22:29.928233  2022-06-29 10:22:35.714741  ...                 3      0.388650  -0.009124          -0.000291          M     5  2015-01-01 ~ 2020-12-31
        2      2  Completed  2022-06-29 08:55:55.268231  2022-06-29 08:56:03.530062  ...                 2      0.380334  -0.022625          -0.000524          M     5  2015-01-01 ~ 2020-12-31
        3      1  Completed  2022-06-27 18:08:36.775176  2022-06-27 18:09:00.530742  ...                 1      0.879670  -0.074594          -0.000771          Q    10  2010-01-01 ~ 2011-01-01
        """
        return cls._list_job()


class strategy_backtest(PrismBacktestTaskComponent):
    @_validate_args
    def __init__(
        self,
        trade: abstract_prismcomponent._AbstractPrismComponent,
        universe: Union[int, str],
        market_impact: str,
        commission_fee: str,
        short_loan_fee: str,
        risk_free_rate: str,
        margin_rate: str,
        cash_interest_rate: str,
        initial_position_type: str,
        initial_position_value: int,
        benchmark: List[Union[int, str]],
        currency: _CurrencyTypeWithReportTrade,
        market_impact_value: float = None,
        commission_fee_value: float = None,
        short_loan_fee_value: float = None,
        risk_free_rate_value: float = None,
        margin_rate_value: float = None,
        cash_interest_rate_value: float = None,
        currency_hedge: bool = False,
        startdate: str = None,
        enddate: str = None,
        unpaid_dividend=None,
    ):
        assert initial_position_type in ["cash", "portfolio"], "Initial Position Type should be one of: cash, portfolio"
        if not isinstance(benchmark, list):
            benchmark = [benchmark]
        benchmark = parse_portfolios_to_portfolioids(benchmark)
        universeid, _ = parse_universe_to_universeid(universe)

        universe_info = get(f"{URL_UNIVERSES}/{universeid}/info")
        universe_startdate = universe_info["Start Date"].values[0]
        universe_enddate = universe_info["End Date"].values[0]

        universe_period_violated = are_periods_exclusive(universe_startdate, universe_enddate, startdate, enddate)

        if universe_period_violated:
            raise PrismValueError(
                f'Strategy Backtest period should overlap with universe period ({str(universe_startdate).split("T")[0]} ~ {str(universe_enddate).split("T")[0]})'
            )

        return super().__init__(
            trade_dataquery=trade._query,
            universeid=int(universeid),
            market_impact=market_impact,
            market_impact_value=market_impact_value,
            commission_fee=commission_fee,
            commission_fee_value=commission_fee_value,
            short_loan_fee=short_loan_fee,
            short_loan_fee_value=short_loan_fee_value,
            risk_free_rate=risk_free_rate,
            risk_free_rate_value=risk_free_rate_value,
            margin_rate=margin_rate,
            margin_rate_value=margin_rate_value,
            cash_interest_rate=cash_interest_rate,
            cash_interest_rate_value=cash_interest_rate_value,
            initial_position_type=initial_position_type,
            initial_position_value=initial_position_value,
            benchmark=benchmark,
            currency=currency,
            currency_hedge=currency_hedge,
            startdate=startdate,
            enddate=enddate,
            unpaid_dividend=unpaid_dividend,
        )


    @_validate_args
    def run(self, portfolioname: str, jobname: str = None, report: bool = True):
        should_overwrite, err_msg = should_overwrite_portfolio(portfolioname, "constructing")
        if not should_overwrite:
            print(f"{err_msg}")
            return
        component_args = copy.deepcopy(self._query["component_args"])
        universeid = component_args.pop("universeid")
        benchmark = component_args.pop("benchmark")

        market_impact = {
            "model": component_args.pop("market_impact"),
            "model_value": component_args.pop("market_impact_value"),
        }
        commission_fee = {
            "model": component_args.pop("commission_fee"),
            "model_value": component_args.pop("commission_fee_value"),
        }
        short_loan_fee = {
            "model": component_args.pop("short_loan_fee"),
            "model_value": component_args.pop("short_loan_fee_value"),
        }
        risk_free_rate = {
            "model": component_args.pop("risk_free_rate"),
            "model_value": component_args.pop("risk_free_rate_value"),
        }
        margin_rate = {
            "model": component_args.pop("margin_rate"),
            "model_value": component_args.pop("margin_rate_value"),
        }
        cash_interest_rate = {
            "model": component_args.pop("cash_interest_rate"),
            "model_value": component_args.pop("cash_interest_rate_value"),
        }
        initial_position = {
            "position_type": component_args.pop("initial_position_type"),
            "position_value": component_args.pop("initial_position_value"),
        }

        task_params = {
            "portfoliopath": portfolioname + ".ppt",
            "universeid": universeid,
            "trade_value_type": "trade",
            "market_impact": market_impact,
            "commission_fee": commission_fee,
            "short_loan_fee": short_loan_fee,
            "risk_free_rate": risk_free_rate,
            "margin_rate": margin_rate,
            "cash_interest_rate": cash_interest_rate,
            "benchmark": benchmark,
            "initial_position": initial_position,
        }
        task_params.update(component_args)
        query = {
            "component_type": "taskcomponent",
            "componentid": self.componentid,
            "component_args": task_params,
        }

        rescontent = None
        with Loader("Strategy Backtest Running... ") as l:
            try:
                rescontent = post(
                    f"{URL_TASK}/{self.componentid}",
                    params={"jobname": jobname},
                    body=query,
                )
            except:
                l.stop()
                raise PrismTaskError("Strategy Backtest has failed.")
            if rescontent["status"] != "Pending":
                l.stop()
                raise PrismTaskError("Strategy Backtest has failed.")

        print(
            f'{rescontent["message"]}: {rescontent["result"][0]["resulttype"]} is {rescontent["result"][0]["resultvalue"]}'
        )

    @classmethod
    @_validate_args
    def get_result(cls, sbid: Union[list, int], data=True, report=False):
        if isinstance(sbid, int):
            sbid = [sbid]
        return cls._get_task_result(cls.__class__.__name__, "Strategy Backtest", sbid, data, report)



import requests
from typing import Union

from ..._common.config import *
from ..._common.const import *
from ._list import _list_dataitem
from .preference import get_quiet_mode
from ..._utils import (
    _validate_args,
    get,
    post,
    patch,
    delete,
    get_sm_attributevalue,
    _process_fileresponse,
    _authentication,
    _fetch_and_parse,
    plot_tree,
    get_sm_attributevalue,
    are_periods_exclusive,
)
from ..._utils.exceptions import PrismNotFoundError, PrismValueError


__all__ = ["list_portfolios", "get_portfolio", "rename_portfolio", "delete_portfolio", "save_index_as_portfolio"]


@_validate_args
def list_portfolio(search: str = None, tree=False):
    """
    Return the all usable portfolios.

    Parameters
    ----------
        search : str, default None
            | Search word for portfolio name, the search is case-insensitive.
            | If None, the entire list is returned.

        tree : bool, default False
            | If True, the folder structure of the portfolio is visualized in a UNIX-style.

    Returns
    -------
        bool or pandas.DataFrame
            | If tree is True, print file system tree of portfolios.
            | If tree is False, return usable portfolios in DataFrame.
            | Columns:

            - *portfolioid*
            - *portfolioname*
            - *portfoliotype*
            - *startdate*
            - *enddate*

    Examples
    --------
        >>> prism.list_portfolio()
        portfolioid  portfolioname  portfoliotype   startdate     enddate
        0            1        S&P 500          index  1700-01-01  2199-12-31
        1            2            ROA       backtest  1700-01-01  2199-12-31
    """

    portfolio_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt", search)
    if portfolio_df.empty:
        raise PrismNotFoundError("No Portfolios found.")
    if not tree:
        return portfolio_df
    plot_tree(portfolio_df["portfolioname"].tolist(), "portfolio")


@_validate_args
def get_portfolio(portfolio, startdate: str = None, enddate: str = None, shownid: list = None):
    """
    Return the specified portfolio in dictionary of dataframes. The dictionary contains Portfolio Level and Portfolio Value.

    Parameters
    ----------
        portfolio : str or int
            | Portfolio name (str) or portfolio id (int) to get.

        startdate : str, default None
            | Start date of the data. The data includes start date.
            | If None, the start date of the portfolio is used. The same applies when the input is earlier than the portfolio start date.

        enddate : str, default None
            | End date of the data. The data excludes end date.
            | If None, the end date of the portfolio is used. The same applies when the input is later than the portfolio end date.

        shownid : list, default None
            | List of Security Master attributes to display with the data.
            | If None, default attributes set in preferences is shown.
            | If empty list ([]), no attribute is shown.

    Returns
    -------
        dict of pandas.DataFrame
            | Portfolios.
            | Keys:

            - *portfoliovalue*
            - *portfoliolevel*

    Examples
    --------
        >>> prism.list_portfolio()
        portfolioid  portfolioname  portfoliotype   startdate     enddate
        0            1        S&P 500          index  1700-01-01  2199-12-31
        1            2            ROA       backtest  1700-01-01  2199-12-31

        >>> prism.get_portfolio(portfolio="S&P 500")
        {
            'portfoliolevel':              date  leveltypeid        value
                            0      1928-01-03            2    17.760000
                            1      1928-01-04            2    17.720000
                            2      1928-01-05            2    17.550000
                            3      1928-01-06            2    17.660000
                            4      1928-01-07            2    17.680000
                            ...           ...          ...          ...
                            38521  2021-03-23            1  8106.832957
                            38522  2021-03-24            1  8062.996308
                            38523  2021-03-25            1  8105.511314
                            38524  2021-03-26            1  8240.381492
                            38525  2021-03-29            1  8233.244940,
            'portfoliovalue':         listingid        date        value
                            0         90036394  1990-09-20   37654000.0
                            1         90036394  1990-09-21   37654000.0
                            2         90036394  1990-09-24   37654000.0
                            3         90036394  1990-09-25   37654000.0
                            4         90036394  1990-09-26   37654000.0
                            ...            ...         ...          ...
                            4062034  706171023  2021-03-24  377861000.0
                            4062035  706171023  2021-03-25  377861000.0
                            4062036  706171023  2021-03-26  377861000.0
                            4062037  706171023  2021-03-29  377861000.0
                            4062038  706171023  2021-03-30  377861000.0
        }
    """

    portfolioid = parse_portfolios_to_portfolioids([portfolio])[0]
    portfolio_info = get(f"{URL_PORTFOLIOS}/{portfolioid}/info")
    portfolio_startdate = portfolio_info["Start Date"].values[0]
    portfolio_enddate = portfolio_info["End Date"].values[0]

    portfolio_period_violated = are_periods_exclusive(portfolio_startdate, portfolio_enddate, startdate, enddate)

    if portfolio_period_violated:
        raise PrismValueError(
            f'Portfolio query period should overlap with portfolio period ({str(portfolio_startdate).split("T")[0]} ~ {str(portfolio_enddate).split("T")[0]})'
        )

    default_shownid = True
    if (shownid is not None) and (len(shownid) == 0):
        shownid = None
        default_shownid = False
    if shownid is not None:
        shownid = [get_sm_attributevalue(a) for a in shownid]

    params = {
        "startdate": startdate,
        "enddate": enddate,
        "shownid": shownid,
        "default_shownid": default_shownid,
    }
    headers = _authentication()
    res = requests.get(URL_PORTFOLIOS + f"/{portfolioid}", params=params, headers=headers)
    ret, _ = _process_fileresponse(res, res.content)
    return list(ret.values())[0]


@_validate_args
def save_index_as_portfolio(
    dataitemid: int,
    startdate: str = None,
    enddate: str = None,
    portfolioname: str = None,
):
    """
    Create a new portfolio containing the constituents of an index.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the index.

        startdate : str, default None
            | Start date of the portfolio. The data includes start date.
            | If None, the earliest available date of the index is used. The same applies when the input is earlier than the earliest available date of the index.

        enddate : str, default None
            | End date of the portfolio. The data excludes end date.
            | If None, the latest available date of the index is used. The same applies when the input is later than the latest available date of the index.

        universename : str, default None
            | Name of the portfolio to be saved from the index.
            | If None, the index name is used.


    Returns
    -------
        status: dict
            Status of save_index_as_portfolio.

    Examples
    --------
        >>> prism.index.portfolio_dataitems("S&P 500")
            dataitemid                                       dataitemname    datamodule       package
        0       4006682                                            S&P 500  S&P US Index  S&P US Index
        1       4006683           S&P 500 - Alternative Carriers (Sub Ind)  S&P US Index  S&P US Index
        2       4006684                 S&P 500 - Biotechnology (Industry)  S&P US Index  S&P US Index
        3       4006685                  S&P 500 - Biotechnology (Sub Ind)  S&P US Index  S&P US Index
        4       4006686                   S&P 500 - Broadcasting (Sub Ind)  S&P US Index  S&P US Index
        ...	        ...	                                               ...           ...           ...
        308     4006990                  S&P 500 Water Utilities (Sub Ind)  S&P US Index  S&P US Index
        309     4006991  S&P 500 Wireless Telecommunication Services (I...  S&P US Index  S&P US Index
        310     4006992  S&P 500 Wireless Telecommunication Services (S...  S&P US Index  S&P US Index
        311     4006993                      S&P 500 Oil (Composite) Index  S&P US Index  S&P US Index
        312     4006994  S&P 500 Semiconductors (Sub Ind)(30-APR-03) In...  S&P US Index  S&P US Index

        >>> prism.save_index_as_portfolio(dataitemid=4006682)
        {'status': 'Success',
        'message': 'Portfolio saved',
        'result': [{'resulttype': 'portfolioid', 'resultvalue': 1}]}

        >>> prism.list_portfolio()
        portfolioid  portfolioname  portfoliotype   startdate     enddate
        0            1        S&P 500          index  1700-01-01  2199-12-31
    """
    if portfolioname is None:
        indices = _list_dataitem(
            datacategory = "Index",
            datacomponent = "Index Level",
        )
        portfolioname = indices[indices["dataitemid"] == dataitemid]["dataitemname"].values[0]

    should_overwrite, err_msg = should_overwrite_portfolio(portfolioname, "saving")
    if not should_overwrite:
        return err_msg
    params = {
        "dataitemid": dataitemid,
        "startdate": startdate,
        "enddate": enddate,
        "path": portfolioname + ".ppt",
    }
    ret = post(URL_PORTFOLIOS + f"/index", params, None)
    print(f'{ret["message"]}: {ret["result"][0]["resulttype"]} is {ret["result"][0]["resultvalue"]}')
    return ret


@_validate_args
def rename_portfolio(old: Union[str, int], new: str):
    """
    Rename portfolio. Location of portfolio within the folder structure can be changed using the method.

    Parameters
    ----------
        old : str
            Name of the existing portfolio to be renamed.

        new : str
            Name of the portfolio after rename.

    Returns
    -------
        status: dict
            Status of rename_portfolio.

    Examples
    --------
        >>> prism.list_portfolio()
        portfolioid                portfolioname  portfoliotype   startdate     enddate
        0            1  Korea Stock Price 200 Index          index  1700-01-01  2199-12-31
        1            2                      S&P 500          index  1700-01-01  2199-12-31
        2            3    Russell 3000 Growth Index          index  1700-01-01  2199-12-31
        3            4           Russell 3000 Index          index  1700-01-01  2199-12-31

        >>> prism.rename_portfolio(old='S&P 500', new='newname_snp')
        {'status': 'Success',
        'message': 'Portfolio renamed',
        'result': [{'resulttype': 'portfolioid', 'resultvalue': 2}]}

        >>> prism.list_portfolio()
        portfolioid                portfolioname  portfoliotype   startdate     enddate
        0            1  Korea Stock Price 200 Index          index  1700-01-01  2199-12-31
        1            2                  newname_snp          index  1700-01-01  2199-12-31
        2            3    Russell 3000 Growth Index          index  1700-01-01  2199-12-31
        3            4           Russell 3000 Index          index  1700-01-01  2199-12-31
    """
    should_overwrite, err_msg = should_overwrite_portfolio(new, "renaming")
    if not should_overwrite:
        return err_msg
    portfolioid = parse_portfolios_to_portfolioids([old])[0]
    ret = patch(URL_PORTFOLIOS + f"/{portfolioid}", {"newpath": new + ".ppt"}, None)
    print(ret["message"])
    return ret


@_validate_args
def delete_portfolio(portfolio: Union[str, int]):
    """
    Delete portfolio.

    Parameters
    ----------
        portfolio : str or int
            Portfolio name (str) or portfolio id (int) to delete.


    Returns
    -------
        status: dict
            Status of delete_portfolio.

    Examples
    --------
        >>> prism.list_portfolio()
        portfolioid  portfolioname  portfoliotype   startdate     enddate
        0            1        S&P 500          index  1700-01-01  2199-12-31
        1            2            ROA       backtest  1700-01-01  2199-12-31

        >>> prism.delete_portfolio(portfolio='S&P 500')
        {'status': 'Success',
        'message': 'Universe deleted',
        'result': [{'resulttype': 'portfolioid', 'resultvalue': 1}]}

        >>> prism.list_portfolio()
        portfolioid  portfolioname  portfoliotype   startdate     enddate
        0            2            ROA       backtest  1700-01-01  2199-12-31
    """
    portfolioid = parse_portfolios_to_portfolioids([portfolio])[0]
    ret = delete(URL_PORTFOLIOS + f"/{portfolioid}")
    print(ret["message"])
    return ret


def parse_portfolios_to_portfolioids(portfolios: list):
    portfolios_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt")
    if portfolios_df.empty:
        raise PrismNotFoundError("No portfolio Found.")

    portfolionames = []
    portfolioids = []
    for p in portfolios:
        if isinstance(p, int):
            portfolioids.append(p)
        elif isinstance(p, str):
            portfolionames.append(p)
        else:
            raise PrismValueError("Please provide portfolio name or portfolio id for portfolio.")

    ret_portfolioids = set()
    if len(portfolionames) != 0:
        matching_portfolios_df = portfolios_df[portfolios_df["portfolioname"].isin(portfolionames)]
        if len(matching_portfolios_df) != len(portfolionames):
            not_matched = list(set(portfolionames) - set(matching_portfolios_df["portfolioname"].tolist()))
            raise PrismNotFoundError(f"No portfolio matching: {not_matched}")
        ret_portfolioids.update(matching_portfolios_df["portfolioid"].tolist())

    if len(portfolioids) != 0:
        matching_portfolios_df = portfolios_df[portfolios_df["portfolioid"].isin(portfolioids)]
        if len(matching_portfolios_df) != len(portfolioids):
            not_matched = list(set(portfolioids) - set(matching_portfolios_df["portfolioid"].tolist()))
            raise PrismNotFoundError(f"No portfolio matching: {not_matched}")
        ret_portfolioids.update(matching_portfolios_df["portfolioid"].tolist())
    return list(ret_portfolioids)


def should_overwrite_portfolio(portfolioname, operation):
    if get_quiet_mode():
        return True, None
    portfolios_df = _fetch_and_parse(URL_PORTFOLIOS, "ppt")
    if portfolios_df.empty:
        return True, None

    if portfolioname in portfolios_df["portfolioname"].to_list():
        overwrite = input(f"{portfolioname} already exists. Do you want to overwrite? (Y/N) \n")
        if overwrite.lower() == "y":
            return True, None
        elif overwrite.lower() == "n":
            return False, f"Abort {operation} portfolio."
        else:
            return False, "Please provide a valid input."
    else:
        return True, None



import shutil
from functools import wraps
from pathlib import Path

import pandas as pd

def _get_params(param_dict):
    return {k: v for k, v in param_dict.items() if not (k.startswith("_") or (k == "self"))}


def _req_call(call_type):
    def decorator(func):
        @wraps(func)
        def wrapper(**kwargs):
            func(**kwargs)
            # vars = func.__code__.co_varnames[1:]
            # params = dict(zip(vars[: len(args)], args))
            # params.update(dict(kwargs))
            fn = getattr(call_type, func.__name__)
            obj = kwargs.pop("self")
            return fn(obj, **kwargs)

        return wrapper

    return decorator


class DisplayablePath(object):
    display_filename_prefix_middle = "├──"
    display_filename_prefix_last = "└──"
    display_parent_prefix_middle = "    "
    display_parent_prefix_last = "│   "

    def __init__(self, path, parent_path, is_last):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        if self.parent:
            self.depth = self.parent.depth + 1
        else:
            self.depth = 0

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(list(path for path in root.iterdir() if criteria(path)), key=lambda s: str(s).lower())
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(path, parent=displayable_root, is_last=is_last, criteria=criteria)
            else:
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path):
        return True

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = self.display_filename_prefix_last if self.is_last else self.display_filename_prefix_middle

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(self.display_parent_prefix_middle if parent.is_last else self.display_parent_prefix_last)
            parent = parent.parent

        return "".join(reversed(parts))


def create_files(paths, directory):
    for path in paths:
        output_file = Path(f"./tmp/{directory}/" + path)
        output_file.parent.mkdir(exist_ok=True, parents=True)
        output_file.write_text("EASTERN STAR")


def plot_tree(paths: list, directory: str):
    shared = [p.split("Shared/")[-1] for p in paths if p.startswith("Shared")]
    non_shared = [p for p in paths if not p.startswith("Shared")]
    shared_rootdir = f"./tmp/Shared"
    non_shared_rootdir = f"./tmp/{directory}"

    create_files(shared, "Shared")
    create_files(non_shared, directory)

    if shared:
        shared_paths = DisplayablePath.make_tree(Path(shared_rootdir))
        for path in shared_paths:
            print(path.displayable())
        shutil.rmtree(shared_rootdir)
    if non_shared:
        non_shared_paths = DisplayablePath.make_tree(Path(non_shared_rootdir))
        for path in non_shared_paths:
            print(path.displayable())
        shutil.rmtree(non_shared_rootdir)
    return

def are_periods_exclusive(startdate1, enddate1, startdate2=None, enddate2=None):
    violated_condition = []
    if (startdate2 is not None) or (enddate2 is not None):
        if enddate2 is not None:
            violated_condition.append(startdate1 > pd.to_datetime(enddate2))
        if startdate2 is not None:
            violated_condition.append(enddate1 < pd.to_datetime(startdate2))
    return any(violated_condition)



from ..._common.const import (
    FinancialPeriodType as _PeriodType,
    FinancialPreliminaryType as _FinancialPreliminaryType,
    CurrencyTypeWithReportTrade as _CurrencyTypeWithReportTrade,
)
from .._req_builder import _list_dataitem
from ..._prismcomponent.prismcomponent import _PrismFinancialComponent, _PrismDataComponent
from ..._utils import _get_params, _validate_args
from ..._utils.exceptions import PrismValueError


__all__ = [
    "balance_sheet",
    "cash_flow",
    "dps",
    # "date",
    "eps",
    "income_statement",
    "segment",
    "ratio",
    "commitment",
    "pension",
    "option",
    "dataitems",
]


_data_category = __name__.split(".")[-1]


class _PrismFinancialDataComponent(_PrismDataComponent, _PrismFinancialComponent):
    _component_category_repr = _data_category

    @classmethod
    def _dataitems(cls, search : str = None, package : str = None):
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )


class balance_sheet(_PrismFinancialDataComponent):
    """
    | Data that pertains to a balance sheet portion in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.financial.balance_sheet.dataitems('asset')
            dataitemid                                       dataitemname
        0       100003                           Trading Asset Securities
        1       100012              Finance Division Other Current Assets
        2       100013                        Other Current Assets, Total
        3       100015                       Deferred Tax Assets, Current
        4       100017                               Other Current Assets
        ..         ...                                                ...
        10      100033                                      Assets, Total
        ..         ...                                                ...
        59      100349  Right-of-Use Assets - Operating Lease - Accumu...
        60      100350      Right-of-Use Assets - Operating Lease - Gross
        61      100351        Right-of-Use Assets - Operating Lease - Net
        62      100365                           Trading Asset Securities
        63      100366                           Trading Portfolio Assets

        >>> ta = prism.financial.balance_sheet(dataitemid=100033, periodtype='Q')
        >>> ta_df = ta.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> ta_df
              listingid        date  currency      period  Assets, Total   Ticker
        0      20108718  2013-05-22       KRW  2013-03-31   9.001210e+11  A004430
        1      20108718  2013-08-21       KRW  2013-06-30   9.132100e+11  A004430
        2      20108718  2013-11-25       KRW  2013-09-30   8.710300e+11  A004430
        3      20108718  2014-03-25       KRW  2013-12-31   8.660220e+11  A004430
        4      20108718  2014-05-15       KRW  2014-03-31   8.890930e+11  A004430
        ...         ...         ...       ...         ...            ...      ...
        4169  278631846  2015-03-31       KRW  2014-12-31   9.511431e+12  A028260
        4170  278631846  2015-05-15       KRW  2015-03-31   8.389397e+12  A028260
        4171  278631846  2015-08-17       KRW  2015-06-30   8.772212e+12  A028260
        4172  278631846  2015-10-15       KRW  2015-09-30   3.926600e+13  A028260
        4173  278631846  2015-11-16       KRW  2015-09-30   3.926658e+13  A028260
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        if periodtype in ["YTD", "LTM"]:
            raise PrismValueError(
                f"Balance Sheet cannot take {periodtype} as periodtype.",
                valid_list=_PeriodType,
                invalids=["YTD", "LTM", "NTM"],
            )
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search : str = None, package : str = None):
        """
        Usable data items for the balance sheet data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.
            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*


        Examples
        --------
            >>> prism.financial.balance_sheet.dataitems('asset')
                dataitemid  ...                                dataitemdescription
            0       100003  ...  This item represents both debt and equity secu...
            1       100012  ...  This item represents all current assets of a f...
            2       100013  ...                                               None
            3       100015  ...  This item represents the deferred tax conseque...
            4       100017  ...  This item represents current assets other than...
            ..         ...  ...                                                ...
            59      100349  ...                                               None
            60      100350  ...                                               None
            61      100351  ...                                               None
            62      100365  ...                                    Mapped from TAS
            63      100366  ...                                    Mapped from TAP
        """
        return cls._dataitems(search=search, package=package)


class cash_flow(_PrismFinancialDataComponent):
    """
    | Data that pertains to a cash flow statement portion in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.financial.cash_flow.dataitems('free cash')
        >>> di[['dataitemid', 'dataitemname']]
           dataitemid              dataitemname
        0      100506    Free Cash Flow / Share
        1      100513    Levered Free Cash Flow
        2      100544  Unlevered Free Cash Flow

        >>> fps = prism.financial.cash_flow(dataitemid=100506, periodtype='LTM')
        >>> fps_df = fps.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> fps_df
              listingid        date currency      period  Free Cash Flow / Share   Ticker
        0      20108718  2011-03-31      KRW  2010-12-31             -450.948236  A004430
        1      20108718  2012-03-30      KRW  2011-12-31             -585.166666  A004430
        2      20108718  2013-01-15      KRW  2012-12-31             2882.006531  A004430
        3      20108718  2013-03-05      KRW  2012-12-31             2881.166666  A004430
        4      20108718  2013-05-22      KRW  2013-03-31             2087.633126  A004430
        ...         ...         ...      ...         ...                     ...      ...
        4044  277591793  2015-11-16      KRW  2015-09-30           -34806.994056  A000030
        4045  278631846  2015-03-31      KRW  2014-12-31             -553.762978  A028260
        4046  278631846  2015-05-15      KRW  2015-03-31             -818.569872  A028260
        4047  278631846  2015-08-17      KRW  2015-06-30             -221.831704  A028260
        4048  278631846  2015-11-16      KRW  2015-09-30             -705.329630  A028260
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the cash flow statement data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*


        Examples
        --------
            >>> prism.financial.cash_flow.dataitems('free cash')
            dataitemid  ...       dataitemdescription
            0      100506  ...                      None
            1      100513  ...    Levered Free Cash Flow
            2      100544  ...  Unlevered Free Cash Flow
        """
        return cls._dataitems(search=search, package=package)


class dps(_PrismFinancialDataComponent):
    """
    | Dividend per share related data.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.financial.dps.dataitems()
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       100547                       Distributable Cash Per Share
        1       100548             Distributable Cash Per Share (Diluted)
        2       100549                                 Dividend Per Share
        3       100550                         Dividend Per Share Class A
        4       100551                         Dividend Per Share Class B
        5       100552                         Special Dividend Per Share
        6       100553         Special Dividend Per Share - Non-Recurring
        7       100554             Special Dividend Per Share - Recurring
        8       100555                 Special Dividend Per Share Class A
        9       100556  Special Dividend Per Share Class A - Non-Recur...
        10      100557     Special Dividend Per Share Class A - Recurring
        11      100558                 Special Dividend Per Share Class B
        12      100559  Special Dividend Per Share Class B - Non-Recur...
        13      100560     Special Dividend Per Share Class B - Recurring

        >>> dps = prism.financial.dps(dataitemid=100549, periodtype='LTM')
        >>> dps_df = dps.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> dps_df
              listingid        date  currency      period  Dividend Per Share   Ticker
        0      20108718  2011-03-31       KRW  2010-12-31                60.0  A004430
        1      20108718  2012-03-30       KRW  2011-12-31                60.0  A004430
        2      20108718  2013-03-05       KRW  2012-12-31                80.0  A004430
        3      20108718  2013-05-22       KRW  2013-03-31                80.0  A004430
        4      20108718  2013-08-21       KRW  2013-06-30                80.0  A004430
        ...         ...         ...       ...         ...                 ...      ...
        3567  277591793  2015-05-15       KRW  2015-03-31               500.0  A000030
        3568  277591793  2015-06-03       KRW  2015-03-31               500.0  A000030
        3569  277591793  2015-08-17       KRW  2015-06-30               750.0  A000030
        3570  277591793  2015-10-29       KRW  2015-09-30               750.0  A000030
        3571  277591793  2015-11-16       KRW  2015-09-30               750.0  A000030
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the dividend per share data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> dpsdi = prism.financial.dps.dataitems()
            >>> dpsdi[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0       100547                       Distributable Cash Per Share
            1       100548             Distributable Cash Per Share (Diluted)
            2       100549                                 Dividend Per Share
            3       100550                         Dividend Per Share Class A
            4       100551                         Dividend Per Share Class B
            5       100552                         Special Dividend Per Share
            6       100553         Special Dividend Per Share - Non-Recurring
            7       100554             Special Dividend Per Share - Recurring
            8       100555                 Special Dividend Per Share Class A
            9       100556  Special Dividend Per Share Class A - Non-Recur...
            10      100557     Special Dividend Per Share Class A - Recurring
            11      100558                 Special Dividend Per Share Class B
            12      100559  Special Dividend Per Share Class B - Non-Recur...
            13      100560     Special Dividend Per Share Class B - Recurring
        """
        return cls._dataitems(search=search, package=package)


# class date(_PrismFinancialDataComponent):
#     """
#     | Relavent dates in the financial statement.
#     | Default frequency is aperiodic daily.

#     Parameters
#     ----------
#         periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
#             | Financial Period in which the financial statement results are reported.
#             | A Financial Period can be of one of the following Period Types:

#             - Annual period (A)
#             - Quarterly period (Q)
#             - Last twelve months (LTM)
#             - Year-to-date (YTD)
#             - Semi-Annual (SA)

#     Returns
#     -------
#         prism._PrismComponent
#             =======================     =====================================================================================================================================================================================================================================================================================
#             Date Type                   Description
#             =======================     =====================================================================================================================================================================================================================================================================================
#             Press Release               Preliminary earnings release. This information is usually released by the company prior to the official filing or release of data.
#             Original                    Original company filing for period. These numbers were the originally filed numbers (not a press release) for this period. In the U.S., this would be represented by a 10K or 10Q SEC filing.
#             Restated                    Results are fundamentally different from the original, i.e., Net Income, Retained Earnings, Total Assets or Cash from Operations are different. Restatements usually happen after an acquisition, divestiture, merger or accounting change.
#             No Change from Original     Appearing again in a later filing, but unchanged from original, or not comparable due to different reporting currencies. These numbers were from a subsequent filing and were recollected but they do not represent changes in the financials that would be considered a restatement.
#             Reclassified                Results somewhat different from original, but bottom line results are the same.
#             =======================     =====================================================================================================================================================================================================================================================================================

#     Examples
#     --------
#         >>> fdate = prism.financial.date('Q')
#         >>> fdate_df = fdate.get_data(universe=1, startdate='2018-01-01')
#         >>> fdate_df
#                  listingid        date      period                 datetype
#         0         20224413  2016-10-27  2015-09-30             Reclassified
#         1         20224413  2016-10-27  2016-09-30                 Original
#         2         20224413  2017-02-08  2015-12-31  No Change from Original
#         3         20224413  2017-02-08  2016-12-31            Press Release
#         4         20224413  2017-03-10  2015-12-31             Reclassified
#         ...            ...         ...         ...                      ...
#         433059  1778113170  2022-09-13  2022-06-30                 Original
#         433060  1786609462  2022-06-07  2021-03-31                 Original
#         433061  1786609462  2022-06-07  2022-03-31                 Original
#         433062  1786609462  2022-08-23  2021-06-30                 Original
#         433063  1786609462  2022-08-23  2022-06-30                 Original
#     """
#     @_validate_args
#     def __init__(
#         self,
#         periodtype: _PeriodType,
#         package : str = None,
#     ):
#         super().__init__(**_get_params(vars()))


class eps(_PrismFinancialDataComponent):
    """
    | Earnings per share related data.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None
    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.financial.eps.dataitems()
        >>> di[['dataitemid', 'dataitemname']]
            dataitemid                                       dataitemname
        0       100561                           Basic Earnings Per Share
        1       100562       Basic Earnings Per Share - Accounting Change
        2       100563   Basic Earnings Per Share - Continuing Operations
        3       100564  Basic Earnings Per Share - Discontinued Operat...
        4       100565     Basic Earnings Per Share - Extraordinary Items
        5       100566  Basic Earnings Per Share - Extraordinary Items...
        6       100567                         Diluted Earnings Per Share
        7       100568     Diluted Earnings Per Share - Accounting Change
        8       100569  Diluted Earnings Per Share - Continuing Operat...
        9       100570  Diluted Earnings Per Share - Discontinued Oper...
        10      100571   Diluted Earnings Per Share - Extraordinary Items
        11      100572  Diluted Earnings Per Share - Extraordinary Ite...
        12      100573                Normalized Basic Earnings Per Share
        13      100574              Normalized Diluted Earnings Per Share
        14      100575                 Reported Basic Earnings Per Share
        15      100576  Reported Basic Earnings Per Share Excl. Extrao...
        16      100577                Reported Diluted Earnings Per Share
        17      100578  Reported Diluted Earnings Per Share Excl. Extr...
        18      100579                                 Revenues Per Share

        >>> eps = prism.financial.eps(dataitemid=100567, periodtype='LTM')
        >>> eps_df = eps.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> eps_df
              listingid        date  currency      period  Diluted Earnings Per Share   Ticker
        0      20108718  2011-03-31       KRW  2010-12-31                  923.751001  A004430
        1      20108718  2012-03-30       KRW  2011-12-31                  418.041666  A004430
        2      20108718  2013-01-15       KRW  2012-12-31                  858.000035  A004430
        3      20108718  2013-03-05       KRW  2012-12-31                  857.750000  A004430
        4      20108718  2013-05-22       KRW  2013-03-31                 1128.355978  A004430
        ...         ...         ...       ...         ...                         ...      ...
        4541  277591793  2015-11-16       KRW  2015-09-30                  764.230312  A000030
        4542  278631846  2015-03-31       KRW  2014-12-31                 4284.419320  A028260
        4543  278631846  2015-05-15       KRW  2015-03-31                  770.272924  A028260
        4544  278631846  2015-08-17       KRW  2015-06-30                  260.679285  A028260
        4545  278631846  2015-11-16       KRW  2015-09-30                23916.737894  A028260
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the earnings per share data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> epsdi = prism.financial.eps.dataitems()
            >>> epsdi[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0       100561                           Basic Earnings Per Share
            1       100562       Basic Earnings Per Share - Accounting Change
            2       100563   Basic Earnings Per Share - Continuing Operations
            3       100564  Basic Earnings Per Share - Discontinued Operat...
            4       100565     Basic Earnings Per Share - Extraordinary Items
            5       100566  Basic Earnings Per Share - Extraordinary Items...
            6       100567                         Diluted Earnings Per Share
            7       100568     Diluted Earnings Per Share - Accounting Change
            8       100569  Diluted Earnings Per Share - Continuing Operat...
            9       100570  Diluted Earnings Per Share - Discontinued Oper...
            10      100571   Diluted Earnings Per Share - Extraordinary Items
            11      100572  Diluted Earnings Per Share - Extraordinary Ite...
            12      100573                Normalized Basic Earnings Per Share
            13      100574              Normalized Diluted Earnings Per Share
            14      100575                 Reported Basic Earnings Per Share
            15      100576  Reported Basic Earnings Per Share Excl. Extrao...
            16      100577                Reported Diluted Earnings Per Share
            17      100578  Reported Diluted Earnings Per Share Excl. Extr...
            18      100579                                 Revenues Per Share
        """
        return cls._dataitems(search=search, package=package)


class income_statement(_PrismFinancialDataComponent):
    """
    | Data that pertains to a income statement portion in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the balance sheet value (Revenue, Expense, etc.)

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> di = prism.financial.income_statement.dataitems('net income')
            dataitemid                             dataitemname
        0       100637                    Net Income to Company
        1       100639                               Net Income
        2       100644          Other Adjustments to Net Income
        3       100645  Net Income Allocable to General Partner
        4       100646   Net Income to Common Incl. Extra Items
        5       100647   Net Income to Common Excl. Extra Items
        6       100703                       Diluted Net Income
        7       100829                               Net Income
        8       100830               Net Income as per SFAS 123
        9       100831  Net Income from Discontinued Operations
        10      100842                    Normalized Net Income

        >>> ni = prism.financial.income_statement(dataitemid=100639, periodtype='LTM')
        >>> ni_df = ni.get_data(universe=1, startdate='2010-01-01', enddate='2015-12-31', shownid=['ticker'])
        >>> ni_df
              listingid        date currency      period    Net Income   Ticker
        0      20108718  2011-03-31      KRW  2010-12-31  2.217002e+10  A004430
        1      20108718  2012-03-30      KRW  2011-12-31  1.003300e+10  A004430
        2      20108718  2013-01-15      KRW  2012-12-31  2.058600e+10  A004430
        3      20108718  2013-03-05      KRW  2012-12-31  2.058600e+10  A004430
        4      20108718  2013-05-22      KRW  2013-03-31  2.714800e+10  A004430
        ...         ...         ...      ...         ...           ...      ...
        4619  278631846  2015-03-31      KRW  2014-12-31  4.556470e+11  A028260
        4620  278631846  2015-05-15      KRW  2015-03-31  8.384555e+10  A028260
        4621  278631846  2015-08-17      KRW  2015-06-30  2.902545e+10  A028260
        4622  278631846  2015-10-15      KRW  2015-09-30  2.820902e+12  A028260
        4623  278631846  2015-11-16      KRW  2015-09-30  2.822057e+12  A028260
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the income statement data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> prism.financial.income_statement.dataitems('income')
                dataitemid  ...                                dataitemdescription
            0       100586  ...  This item represents the interest and investme...
            1       100588  ...  This item represents fee from non-fund based a...
            2       100607  ...  This item represents all other operating expen...
            3       100610  ...  This item represents the difference between th...
            4       100612  ...  This item represents the interest and investme...
            ..         ...  ...                                                ...
            71      100902  ...                                               None
            72      100903  ...  This item represents the total sub-lease incom...
            73      100904  ...  This item represents Taxes other than excise a...
            74      100905  ...  This item represents all taxes other than inco...
            75      100906  ...  This item represents refund of any tax amount ...
        """
        return cls._dataitems(search=search, package=package)


class segment(_PrismFinancialDataComponent):
    """
    | Data that pertains to a specific segment or division within a company.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item.

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> segment = prism.financial.segment.dataitems()
        >>> segment[['dataitemid', 'dataitemname']]
           dataitemid                                     dataitemname
        0      104729                        Business Segments - CAPEX
        1      104730  Business Segments - Depreciation & Amortization
        2      104731                       Business Segments - EBITDA
        3      104732                          Business Segments - EBT
        4      104733                 Business Segments - Gross Profit
        5      104734           Business Segments - Income Tax Expense
        6      104735             Business Segments - Interest Expense
        7      104736                        Business Segments - NOPAT
        8      104737                   Business Segments - Net Income
        >>> seg = prism.financial.segment(104731, periodtype='Q')
        >>> seg.get_data(universe=1, startdate='2010-01-01', enddate='2015-01-01', shownid=['ticker'])
           listingid        date            segment  latestperiod  currency  Business Segments - EBITDA: 0 Quarters Back   Ticker
        0  144523336  2013-01-31  Convenience Store    2012-12-31       KRW                                 4.920000e+10  A007070
        1  144523336  2013-01-31        Supermarket    2012-12-31       KRW                                 6.900000e+09  A007070
        2  144523336  2013-03-29  Convenience Store    2012-12-31       KRW                                 4.920000e+10  A007070
        3  144523336  2013-03-29        Supermarket    2012-12-31       KRW                                 6.900000e+09  A007070
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the segment data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to segment data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> segdi = prism.financial.segment.dataitems()
            >>> segdi[['dataitemid', 'dataitemname']]
            dataitemid                                     dataitemname
            0      104729                        Business Segments - CAPEX
            1      104730  Business Segments - Depreciation & Amortization
            2      104731                       Business Segments - EBITDA
            3      104732                          Business Segments - EBT
            4      104733                 Business Segments - Gross Profit
            5      104734           Business Segments - Income Tax Expense
            6      104735             Business Segments - Interest Expense
            7      104736                        Business Segments - NOPAT
            8      104737                   Business Segments - Net Income
        """
        return cls._dataitems(search=search, package=package)


class ratio(_PrismFinancialDataComponent):
    """
    | Data that pertains to a ratio data in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item.

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> ratdi = prism.financial.ratio.dataitems()
        >>> ratdi[['dataitemid', 'dataitemname']]
             dataitemid                                       dataitemname
        0        104934                       Annualized Dividend Payout %
        1        104935                      Annualized Dividend Per Share
        2        104936                        Annualized Dividend Yield %
        3        104937    3 Yr. Compound Net Capital Expenditure Growth %
        4        104938    5 Yr. Compound Net Capital Expenditure Growth %
        ...         ...                                                ...
        303      112907                        Liabilities / Assets, Total
        304      113691                Net Working Capital / Total Revenue
        305      113692                 Net Working Capital / Total Assets
        306      113695                           Working Capital Turnover
        307      114803  Altman Z Score Using the Average Stock Informa...
        >>> rat = prism.financial.ratio(112907, periodtype='Q')
        >>> rat.get_data(universe=1, startdate='2010-01-01', enddate='2015-01-01', shownid=['ticker'])
               listingid        date  latestperiod  Liabilities / Assets, Total: 0 Quarters Back   Ticker
        0       20216251  2010-01-01    2009-09-30                                       82.1596     None
        1       49782109  2010-01-06    2009-09-30                                       10.7062     None
        2       20168002  2010-01-08    2009-09-30                                       93.1549  A005350
        3       31779033  2010-01-11    2009-09-30                                       37.3517  A050470
        4       31779068  2010-01-12    2009-09-30                                       49.1868  A056080
        ...          ...         ...           ...                                           ...      ...
        22657  224013404  2014-12-31    2014-09-30                                       67.5691  A141070
        22658  237676086  2014-12-31    2014-09-30                                       35.5406  A158380
        22659  248503946  2014-12-31    2014-09-30                                       38.1093  A131970
        22660  248629939  2014-12-31    2014-09-30                                        8.8119  A171120
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the ratio data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to ratio data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> ratdi = prism.financial.segment.dataitems()
            >>> ratdi[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0        104934                       Annualized Dividend Payout %
            1        104935                      Annualized Dividend Per Share
            2        104936                        Annualized Dividend Yield %
            3        104937    3 Yr. Compound Net Capital Expenditure Growth %
            4        104938    5 Yr. Compound Net Capital Expenditure Growth %
            ...         ...                                                ...
            303      112907                        Liabilities / Assets, Total
            304      113691                Net Working Capital / Total Revenue
            305      113692                 Net Working Capital / Total Assets
            306      113695                           Working Capital Turnover
            307      114803  Altman Z Score Using the Average Stock Informa...
        """
        return cls._dataitems(search=search, package=package)


class commitment(_PrismFinancialDataComponent):
    """
    | Data that pertains to a commitment related data (such as operating leases etc.) in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item.

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> comdi = prism.financial.commitment.dataitems()
        >>> comdi[['dataitemid', 'dataitemname']]
             dataitemid                               dataitemname
        0        104830     Capital Lease Payment Due, Current Yr.
        1        104831  Capital Lease Payment Due, Current Yr. +1
        2        104832  Capital Lease Payment Due, Current Yr. +2
        3        104833  Capital Lease Payment Due, Current Yr. +3
        4        104834  Capital Lease Payment Due, Current Yr. +4
        ...         ...                                        ...
        117      500282                     Year 1 - (Annual Only)
        118      500283                     Year 2 - (Annual Only)
        119      500284                     Year 3 - (Annual Only)
        120      500285                     Year 4 - (Annual Only)
        121      500286                     Year 5 - (Annual Only)
        >>> com = prism.financial.commitment(104830, periodtype='Q')
        >>> com.get_data(universe=1, startdate='2010-01-01', enddate='2015-01-01', shownid=['ticker'])
               listingid        date  latestperiod  currency  Capital Lease Payment Due, Current Yr.: 0 Quarters Back   Ticker
        0       20216251  2010-01-01    2009-09-30       KRW                                                      NaN     None
        1       49782109  2010-01-06    2009-09-30       KRW                                                      NaN     None
        2       20168002  2010-01-08    2009-09-30       KRW                                                      NaN  A005350
        3       31779033  2010-01-11    2009-09-30       KRW                                                      NaN  A050470
        4       31779068  2010-01-12    2009-09-30       KRW                                                      NaN  A056080
        ...          ...         ...           ...       ...                                                      ...      ...
        22657  224013404  2014-12-31    2014-09-30       KRW                                                      NaN  A141070
        22658  237676086  2014-12-31    2014-09-30       KRW                                                      NaN  A158380
        22659  248503946  2014-12-31    2014-09-30       KRW                                                      NaN  A131970
        22660  248629939  2014-12-31    2014-09-30       KRW                                                      NaN  A171120
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the commitment data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to commitment data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> comdi = prism.financial.commitment.dataitems()
            >>> comdi[['dataitemid', 'dataitemname']]
                dataitemid                               dataitemname
            0        104830     Capital Lease Payment Due, Current Yr.
            1        104831  Capital Lease Payment Due, Current Yr. +1
            2        104832  Capital Lease Payment Due, Current Yr. +2
            3        104833  Capital Lease Payment Due, Current Yr. +3
            4        104834  Capital Lease Payment Due, Current Yr. +4
            ...         ...                                        ...
            117      500282                     Year 1 - (Annual Only)
            118      500283                     Year 2 - (Annual Only)
            119      500284                     Year 3 - (Annual Only)
            120      500285                     Year 4 - (Annual Only)
            121      500286                     Year 5 - (Annual Only)
        """
        return cls._dataitems(search=search, package=package)


class pension(_PrismFinancialDataComponent):
    """
    | Data that pertains to a pension related data in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item.

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> pendi = prism.financial.pension.dataitems()
        >>> pendi[['dataitemid', 'dataitemname']]
             dataitemid                                       dataitemname
        0        100299                          Minimum Pension Liability
        1        100370    Un-funded Vested Pension Liabilities - Domestic
        2        100371     Un-funded Vested Pension Liabilities - Foreign
        3        100741  Expected Rate of Return on Pension Assets - Do...
        4        100742  Expected Rate of Return on Pension Assets - Fo...
        ...         ...                                                ...
        592      500346              Net Liability/(Asset) - (Annual Only)
        593      500347               Unfunded Liabilities - (Annual Only)
        594      500348     Expected Return on Plan Assets - (Annual Only)
        595      500349       Actual Return on Plan Assets - (Annual Only)
        596      500350  Expected LT Return Rate on Plan Assets (%) - (...
        >>> pen = prism.financial.pension(104830, periodtype='Q')
        >>> pen.get_data(universe=1, startdate='2010-01-01', enddate='2015-01-01', shownid=['ticker'])
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the pension data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to pension data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> pendi = prism.financial.pension.dataitems()
            >>> pendi[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0        100299                          Minimum Pension Liability
            1        100370    Un-funded Vested Pension Liabilities - Domestic
            2        100371     Un-funded Vested Pension Liabilities - Foreign
            3        100741  Expected Rate of Return on Pension Assets - Do...
            4        100742  Expected Rate of Return on Pension Assets - Fo...
            ...         ...                                                ...
            592      500346              Net Liability/(Asset) - (Annual Only)
            593      500347               Unfunded Liabilities - (Annual Only)
            594      500348     Expected Return on Plan Assets - (Annual Only)
            595      500349       Actual Return on Plan Assets - (Annual Only)
            596      500350  Expected LT Return Rate on Plan Assets (%) - (...
        """
        return cls._dataitems(search=search, package=package)


class option(_PrismFinancialDataComponent):
    """
    | Data that pertains to a options and warrants related data in financial statement.
    | Default frequency is aperiodic daily.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item.

        periodtype : str, {'A', 'Annual', 'SA', 'Semi-Annual', 'Quarterly', 'Q', 'LTM', 'YTD'}
            | Financial Period in which the financial statement results are reported.
            | A Financial Period can be of one of the following Period Types:

            - Annual period (A)
            - Quarterly period (Q)
            - Last twelve months (LTM)
            - Year-to-date (YTD)
            - Semi-Annual (SA)

        preliminary : str, {'keep', 'ignore', 'null'}, default 'keep'
            - keep : keep preliminary data
            - ignore : ignore preliminary data

            .. admonition:: Note
                :class: note

                | If the 'ignore' option is chosen, preliminary reports are disregarded entirely, as if they never existed.
                |
                | Consequently, if a revision occurs on the same day as the preliminary report, the latest period (period 0) will continue to display the previous period of preliminary reporting, and it will not be updated until the official report is released.

            - null : nulled-out preliminary data

        currency : str, {'report', 'trade', ISO3 currency}, default 'report'
            | Desired currency for the financial data.

            - report : financial reporting currency for a given listing (i.e for Apple - USD, Tencent - CNY)
            - trade : trading currency for a given listing (i.e for Apple - USD, Tencent - HKD)
            - ISO3 currency : desired currency in ISO 4217 format (i.e USD, EUR, JPY, KRW, etc.)

            .. admonition:: Warning
                :class: warning

                | If a selected data item is not a currency value (i.e airplanes owned), the currency input will be ignored.
                | It will behave like parameter input currency=None

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> optdi = prism.financial.option.dataitems()
        >>> optdi[['dataitemid', 'dataitemname']]
             dataitemid                                       dataitemname
        0        100668                       Average Price paid per Share
        1        100883                  Shares Purchased - Quarter, Total
        2        104885                       Stock Options Exercise Price
        3        104886                          Stock Options Grant Price
        4        104898  Stock Options Outstanding At The Beginning of ...
        ...         ...                                                ...
        109      107688    Exercisable Warrants W / Average Exercise Price
        110      107689  Exercisable Warrants W / Average Remaining Lif...
        111      107690     Outstanding Warrants Aggregate Intrinsic Value
        112      107691     Exercisable Warrants Aggregate Intrinsic Value
        113      107693                             Stock Option Plan Name
        >>> opt = prism.financial.pension(100668, periodtype='Q')
        >>> opt.get_data(universe=1, startdate='2010-01-01', enddate='2015-01-01', shownid=['ticker'])
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
        periodtype: _PeriodType,
        preliminary: _FinancialPreliminaryType = "keep",
        currency: _CurrencyTypeWithReportTrade = "report",
        package : str = None,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable data items for the option data component.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                | Data items that belong to option data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*

        Examples
        --------
            >>> optdi = prism.financial.option.dataitems()
            >>> optdi[['dataitemid', 'dataitemname']]
                dataitemid                                       dataitemname
            0        100668                       Average Price paid per Share
            1        100883                  Shares Purchased - Quarter, Total
            2        104885                       Stock Options Exercise Price
            3        104886                          Stock Options Grant Price
            4        104898  Stock Options Outstanding At The Beginning of ...
            ...         ...                                                ...
            109      107688    Exercisable Warrants W / Average Exercise Price
            110      107689  Exercisable Warrants W / Average Remaining Lif...
            111      107690     Outstanding Warrants Aggregate Intrinsic Value
            112      107691     Exercisable Warrants Aggregate Intrinsic Value
            113      107693                             Stock Option Plan Name
        """
        return cls._dataitems(search=search, package=package)


@_validate_args
def dataitems(search: str = None, package: str = None):
    """
    Usable data items for the financial data category.

    Parameters
    ----------
        search : str, default None
            | Search word for dataitems name, the search is case-insensitive.

        package : str, default None
            | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:

                - *datamodule*
                - *datacomponent*
                - *dataitemid*
                - *datadescription*


    Examples
    --------
    >>> prism.financial.dataitems('revenue')
        dataitemid  ...                                dataitemdescription
    0       100044  ...  This item represents revenues, which are recei...
    1       100053  ...  This item represents revenues which are receiv...
    2       100196  ...  This item represents revenues relating to peri...
    3       100228  ...  This item represents the portion of realized r...
    4       100263  ...  This item represents the bonds which are issue...
    5       100408  ...  This item represents change in revenues relati...
    6       100483  ...                                From AP Tag CFURXNC
    7       100579  ...                                  Revenue Per Share
    8       100580  ...  This item represents the total revenues that a...
    9       100581  ...                                               None
    10      100582  ...  This item represents the revenues generated by...
    11      100583  ...  This item represents revenues generated by the...
    """
    return _list_dataitem(
            datacategoryid=_PrismFinancialDataComponent.categoryid,
            datacomponentid=None,
            search=search,
            package=package,
        )




from .._req_builder import _list_dataitem
from ..._prismcomponent.prismcomponent import _PrismDataComponent, _PrismComponent
from ..._utils import _validate_args, _get_params


__all__ = [
    'social',
    'governance',
    'summary',
    'environmental',
]


_data_category = __name__.split(".")[-1]


class _PrismESGComponent(_PrismComponent, _PrismDataComponent):
    _component_category_repr = _data_category

    @classmethod
    def _dataitems(cls, search : str = None, package : str = None):
        return _list_dataitem(
            datacategoryid=cls.categoryid,
            datacomponentid=cls.componentid,
            search=search,
            package=package,
        )


class environmental(_PrismESGComponent):
    """
    Retrieves detailed criteria scores and weights that constitute the Environmental Dimension evaluation.

    The Environmental Dimension allows you to access to comprehensive information concerning the company's environmental impact and sustainability practices. It examines the company's efforts to mitigate environmental risks, conserve resources, reduce emissions, and promote eco-friendly initiatives.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> esg = prism.esg.environmental(700088)
        >>> esg.get_data("Semiconductor", "2020-01-01", shownid=['Company Name'])
                listingid        date  value          Company Name
        0         2597014  2020-01-17    0.0   CMC Materials, Inc.
        1         2597014  2020-01-17   67.0   CMC Materials, Inc.
        2         2597014  2020-01-17   33.0   CMC Materials, Inc.
        3         2597014  2020-01-17   26.0   CMC Materials, Inc.
        4         2597014  2020-01-17   41.0   CMC Materials, Inc.
        ...           ...         ...    ...                   ...
        36680  1682846833  2023-01-20   44.0  GLOBALFOUNDRIES Inc.
        36681  1682846833  2023-01-20   74.0  GLOBALFOUNDRIES Inc.
        36682  1682846833  2023-01-20   50.0  GLOBALFOUNDRIES Inc.
        36683  1682846833  2023-01-20    0.0  GLOBALFOUNDRIES Inc.
        36684  1682846833  2023-01-20   33.0  GLOBALFOUNDRIES Inc.


    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable dataitems for criteria in Environmental Dimension.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *dataitemid*
                - *dataitemname*
                - *datamodule*
                - *package*
                - *datadescription*


        Examples
        --------
            >>> env_dataitems = prism.esg.environmental.dataitems()
            >>> print(env_dataitems)
                dataitemid                              dataitemname  datamodule                package                                dataitemdescription
            0       700088                      Biodiversity - Score       Score  S&P Global ESG Scores  Score of Biodiversity\nExtractive industries a...
            1       700089                Building Materials - Score       Score  S&P Global ESG Scores  Score of Building Materials\nA substantial per...
            2       700090  Business Risks and Opportunities - Score       Score  S&P Global ESG Scores  Score of Business Risks and Opportunities\nFin...
            3       700091                  Climate Strategy - Score       Score  S&P Global ESG Scores  Score of Climate Strategy\nMost industries are...
            4       700092                     Co-Processing - Score       Score  S&P Global ESG Scores  Score of Co-Processing\nCo-processing involves...
            ..         ...                                       ...         ...                                                                       ...
            65      700233                       Energy Mix - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Energy Mix\nOil & Gas produc...
            66      700234                 Fleet Management - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Fleet Management\nThe airlin...
            67      700235         Sustainable Construction - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Sustainable Construction\nTh...
            68      700236                 Circular Fashion - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Circular Fashion\nThe fashio...
            69      700237         Decarbonization Strategy - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Decarbonization Strategy\nAs...
        """
        return cls._dataitems(search=search, package=package)


class social(_PrismESGComponent):
    """
    Retrieves detailed criteria scores and weights that constitute the Social Dimension evaluation.

    The Social Dimension allows you to access comprehensive information about how the company manages its internal and external relationships, including interactions with employees and the communities where it operates. The provided criteria scores and weights give you a deeper insight into the company's social impact and performance.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> esg = prism.esg.social(700008)
        >>> esg.get_data("Semiconductor", "2020-01-01", shownid=['Company Name'])
                listingid        date  value          Company Name
        0         2597014  2020-01-17    0.0   CMC Materials, Inc.
        1         2597014  2020-01-17   33.0   CMC Materials, Inc.
        2         2597014  2020-01-17   67.0   CMC Materials, Inc.
        3         2597014  2020-01-17   26.0   CMC Materials, Inc.
        4         2597014  2020-01-17   10.0   CMC Materials, Inc.
        ...           ...         ...    ...                   ...
        36680  1682846833  2023-01-20   44.0  GLOBALFOUNDRIES Inc.
        36681  1682846833  2023-01-20   74.0  GLOBALFOUNDRIES Inc.
        36682  1682846833  2023-01-20   50.0  GLOBALFOUNDRIES Inc.
        36683  1682846833  2023-01-20    0.0  GLOBALFOUNDRIES Inc.
        36684  1682846833  2023-01-20   33.0  GLOBALFOUNDRIES Inc.
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable dataitems for criteria in Social Dimension.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *dataitemid*
                - *dataitemname*
                - *datamodule*
                - *package*
                - *datadescription*

        Examples
        --------
            >>> soc_dataitems = prism.esg.social.dataitems()
            >>> print(soc_dataitems)
                dataitemid                                       dataitemname  datamodule                package  dataitemdescription
            0       700008                            Access to Water - Score       Score  S&P Global ESG Scores  Score of Access to Water\nWater is a pre-requi...
            1       700009                     Addressing Cost Burden - Score       Score  S&P Global ESG Scores  Score of Addressing Cost Burden\nDue to aging ...
            2       700010                   Asset Closure Management - Score       Score  S&P Global ESG Scores  Score of Asset Closure Management\nMining acti...
            3       700011                                  Bioethics - Score       Score  S&P Global ESG Scores  Score of Bioethics\nDetailed and transparent p...
            4       700012  Controversial Issues: Dilemmas in Lending & Fi...       Score  S&P Global ESG Scores  Score of Controversial Issues: Dilemmas in Len...
            ..         ...                                                ...         ...                    ...
            69      700155                        Health & Nutrition - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Health & Nutrition\nPoor die...
            70      700156                        Privacy Protection - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Privacy Protection\nNetworke...
            71      700157          Customer Relationship Management - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Customer Relationship Manage...
            72      700158                       Marketing Practices - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Marketing Practices\nAggress...
            73      700159  Sustainable Marketing & Brand Perception - Weight      Weight  S&P Global ESG Scores  Scoring Weight of Sustainable Marketing & Bran...
        """
        return cls._dataitems(search=search, package=package)


class governance(_PrismESGComponent):
    """
    Retrieves detailed criteria scores and weights that constitute the Governance & Economic Dimension evaluation.

    The Governance & Economic Dimension allows you to access comprehensive information about company's leadership, executive pay, audits, internal controls, and shareholder rights and the relations to external stakeholders such as suppliers and customers. The provided criteria scores and weights give you a deeper insight into the company's governance practices and economic performance.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> esg = prism.esg.governance(700045)
        >>> esg.get_data("Semiconductor", "2020-01-01", shownid=['Company Name'])
                listingid        date  value          Company Name
        0         2597014  2020-01-17    0.0   CMC Materials, Inc.
        1         2597014  2020-01-17   33.0   CMC Materials, Inc.
        2         2597014  2020-01-17   26.0   CMC Materials, Inc.
        3         2597014  2020-01-17   67.0   CMC Materials, Inc.
        4         2597014  2020-01-17   10.0   CMC Materials, Inc.
        ...           ...         ...    ...                   ...
        36680  1682846833  2023-01-20   44.0  GLOBALFOUNDRIES Inc.
        36681  1682846833  2023-01-20   74.0  GLOBALFOUNDRIES Inc.
        36682  1682846833  2023-01-20   50.0  GLOBALFOUNDRIES Inc.
        36683  1682846833  2023-01-20   33.0  GLOBALFOUNDRIES Inc.
        36684  1682846833  2023-01-20    0.0  GLOBALFOUNDRIES Inc.
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable dataitems for criteria in Governance & Economic Dimension.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *dataitemid*
                - *dataitemname*
                - *datamodule*
                - *package*
                - *datadescription*


        Examples
        --------
            >>> env_dataitems = prism.esg.environmental.dataitems()
            >>> print(env_dataitems)
                dataitemid                                 dataitemname datamodule  package   dataitemdescription
            0       700045         Anti-crime Policy & Measures - Score      Score  S&P Global ESG Scores  Score of Anti-crime Policy & Measures\nCrimina...
            1       700046                     Antitrust Policy - Score      Score  S&P Global ESG Scores  Score of Antitrust Policy\nAs global competiti...
            2       700047                     Brand Management - Score      Score  S&P Global ESG Scores  Score of Brand Management\nA brand is a living...
            3       700048  Financial Stability & Systemic Risk - Score      Score  S&P Global ESG Scores  Score of Financial Stability & Systemic Risk\n...
            4       700049                      Business Ethics - Score      Score  S&P Global ESG Scores  Score of Business Ethics\nThe criterion evalua...
            ..         ...                                          ...        ...                                                                       ...
            81      700198             Supply Chain Management - Weight     Weight  S&P Global ESG Scores  Scoring Weight of Supply Chain Management\nIn ...
            82      700199            Sustainable Construction - Weight     Weight  S&P Global ESG Scores  Scoring Weight of Sustainable Construction\nTh...
            83      700200                 Sustainable Finance - Weight     Weight  S&P Global ESG Scores  Scoring Weight of Sustainable Finance\nFinanci...
            84      700201                        Tax Strategy - Weight     Weight  S&P Global ESG Scores  Scoring Weight of Tax Strategy\nTax competitio...
            85      700202                    Water Operations - Weight     Weight  S&P Global ESG Scores  Scoring Weight of Water Operations\nWater Oper...
        """
        return cls._dataitems(search=search, package=package)


class summary(_PrismESGComponent):
    """
    Provides the overall ESG score and provide dimension level scores and weights.

    The summary takes in ESG scores and their corresponding weights to calculate the overall Environmental, Social, and Governance (ESG) score for a company. It also provides dimension level scores and weights, offering insights into the company's performance in each ESG dimension.

    Parameters
    ----------
        dataitemid : int
            | Unique identifier for the different data item. This identifies the type of the value.

    Returns
    -------
        prism._PrismComponent

    Examples
    --------
        >>> esg = prism.esg.summary(700001)
        >>> esg.get_data("Semiconductor", "2020-01-01", shownid=['Company Name'])
                listingid        date  value                                     Company Name
        0         2597014  2020-01-17     16                              CMC Materials, Inc.
        1         2597014  2020-01-17     39                              CMC Materials, Inc.
        2         2597014  2020-01-17     14                    Universal Display Corporation
        3         2597014  2020-01-17     30                    Universal Display Corporation
        4       412298450  2020-01-17     18                             Ichor Holdings, Ltd.
        ...           ...         ...    ...                                              ...
        36680    79144374  2023-01-20     84                         WIN Semiconductors Corp.
        36681   140445201  2023-01-20     23                                 SÜSS MicroTec SE
        36682   627069513  2023-01-20     14  Advanced Micro-Fabrication Equipment Inc. China
        36683   627069513  2023-01-20     23  Advanced Micro-Fabrication Equipment Inc. China
        36684  1682846833  2023-01-20     40                             GLOBALFOUNDRIES Inc.
    """
    @_validate_args
    def __init__(
        self,
        dataitemid: int,
    ):
        super().__init__(**_get_params(vars()))

    @classmethod
    @_validate_args
    def dataitems(cls, search: str = None, package: str = None):
        """
        Usable dataitems for summary.

        Parameters
        ----------
            search : str, default None
                | Search word for dataitems name, the search is case-insensitive.

            package : str, default None
                | Search word for package name, the search is case-insensitive.

        Returns
        -------
            pandas.DataFrame
                Data items that belong to cash flow statement data component.

            Columns:
                - *dataitemid*
                - *dataitemname*
                - *datamodule*
                - *package*
                - *datadescription*


        Examples
        --------
            >>> sum_dataitems = prism.esg.summary.dataitems()
            >>> print(sum_dataitems)
            dataitemid              dataitemname  datamodule                package                                dataitemdescription
            0      700001       ESG Overall - Score       Score  S&P Global ESG Scores  The overarching score reflecting a company's p...
            1      700002     Environmental - Score       Score  S&P Global ESG Scores  The Environment dimension-level score that ref...
            2      700003            Social - Score       Score  S&P Global ESG Scores  The Social dimension-level score that reflects...
            3      700004        Governance - Score       Score  S&P Global ESG Scores  The Governance and Economic dimension-level sc...
            4      700005    Environmental - Weight      Weight  S&P Global ESG Scores  The sum of Environmental criteria weights. Thi...
            5      700006           Social - Weight      Weight  S&P Global ESG Scores  The sum of the Social criteria weights. This r...
            6      700007       Governance - Weight      Weight  S&P Global ESG Scores  The sum of the Governance and Economic criteri...
        """
        return cls._dataitems(search=search, package=package)


def dataitems(search: str = None, package: str = None):
    return _list_dataitem(
        datacategoryid=_PrismESGComponent.categoryid,
        datacomponentid=None,
        search=search,
        package=package,
    )



import webbrowser

from prism._utils.auth_utils import _get_document_authentication_token

from ..._common.config import *
from ..._utils import _get_web_authentication_token


__all__ = ['job_manager', 'preference_setting', 'securitymaster_search', 'dataitems_search']


def job_manager():
    """
    Open job manager in a new web browser tab.

    Returns
    -------
        None

    Notes
    -----
        The user might be asked to login if not already logged in.

    Examples
    --------
        >>> import prism
        >>> prism.job_manager()

        .. image:: ../../_static/jobmanager-api-ref-screenshot.png

    """
    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + '/jobmanager?token=' + str(web_auth_token)
    webbrowser.open(link, new=2)


def preference_setting():
    """
    Open preference setting window on a new browser tab.

    Returns
    -------
        None

    Notes
    -----
        The user will be asked to login if not already logged in.

    Examples
    --------
        >>> import prism
        >>> prism.preference_setting()
    """

    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + '/preference?token=' + str(web_auth_token)
    webbrowser.open(link, new=2)


def securitymaster_search():
    """
    Opens the security master search window on a new browser tab.

    Notes
    -----
        This function opens the security master search page on the Prism web application in a new browser tab.
        The user will be asked to login if they are not already logged in.

    Returns
    -------
        None
    """
    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + '/securitymaster?token=' + str(web_auth_token)
    webbrowser.open(link, new=2)


def dataitem_search():
    """
    Opens the data item search window on a new browser tab.

    Notes
    -----
        This function opens the data item search page on the Prism web application in a new browser tab.
        The user will be asked to login if they are not already logged in.

    Returns
    -------
        None
    """
    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + '/dataitems?token=' + str(web_auth_token)
    webbrowser.open(link, new=2)


def finder():
    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + '/finder?token=' + str(web_auth_token)
    webbrowser.open(link, new=2)


def finder(file_type='dataquery'):
    """
    | Open query manager window on a new browser tab.
    | Users are able to:

    1. Share files or a folder within the same account.
    2. File operations, such as Copy, Cut, Paste, Delete, Move etc..
    3. Create, Delete Folders.
    4. Check properties of a file.

    Parameters
    ----------
        file_type: str {'dataquery', 'taskquery', 'universe', 'portfolio'}, default 'dataquery'
            Open correspoding type of file in browser tab.

    Returns
    -------
        None
            .. admonition:: Warning
                :class: warning

                The user will be asked to login if not already logged in.

    Examples
    --------
        >>> prism.finder()

        .. image:: ../../_static/finder.png
    """
    web_auth_token = _get_web_authentication_token()
    link = ROOT_EXT_WEB_URL + f'/finder?type={file_type}&token={web_auth_token}'
    webbrowser.open(link, new=2)


def open_document(version: str = 'latest'):
    doc_auth_token = _get_document_authentication_token()
    link = ROOT_DOCUMENT_URL + '/' + version + '/?token=' + str(doc_auth_token)
    webbrowser.open(link, new=2)
